﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;

namespace Plux.Framework {
  [Extension]
  [Plug("Plux.LayoutManager")]
  public class LayoutManager : ExtensionBase, ILayoutManager1 {
    internal static readonly Dictionary<PlugInfo, LayoutInfo> layoutInfos
        = new Dictionary<PlugInfo, LayoutInfo>();

    private readonly List<ILayoutPanel> layoutPanels = new List<ILayoutPanel>();

    public LayoutManager() {}

    public LayoutManager(SlotInfo layoutSlot) : this(layoutSlot, string.Empty, string.Empty) {}

    public LayoutManager(SlotInfo layoutSlot, string layoutPanelTag, string elementTag) {
      LayoutPanelTag = layoutPanelTag;
      ElementTag = elementTag;
      LayoutSlot = layoutSlot;
    }

    public string LayoutPanelTag { get; set; }
    public string ElementTag { get; set; }

    //TODO: WORKS ONLY IF SLOT MULTIPLE = TURE

    private SlotInfo layoutSlot;
    private PlugTypeInfo panelPlugType;

    private bool wasLayoutSlotAutoPlug;
    private bool wasLayoutSlotAutoRelease;
    //private bool wasLayoutSlotMultiple;
    private bool wasLayoutSlotOpen;

    public SlotInfo LayoutSlot {
      get { return layoutSlot; }
      set {
        if (layoutSlot == value) {
          return;
        }
        if (layoutSlot != null) {
          Deactivate();
          layoutSlot.AutoPlug = wasLayoutSlotAutoPlug;
          layoutSlot.AutoRelease = wasLayoutSlotAutoRelease;
          if (!wasLayoutSlotOpen) {
            layoutSlot.Close();
          }
        }
        layoutSlot = value;
        if (layoutSlot == null) {
          PanelPlugType = null;
          return;
        }
        wasLayoutSlotAutoPlug = layoutSlot.AutoPlug;
        wasLayoutSlotAutoRelease = layoutSlot.AutoRelease;
        //wasLayoutSlotMultiple = slot.Multiple;
        wasLayoutSlotOpen = layoutSlot.IsOpen;
        layoutSlot.AutoPlug = false;

        if (PanelPlugType == null || PanelPlugType.Name != layoutSlot.Name) {
          PanelPlugType = FindPanelPlugType(Runtime.Repository.PlugTypeInfos.Find(layoutSlot.Name));
        } else {         
          Activate();
        }
      }
    }

    private PlugTypeInfo PanelPlugType {
      get { return panelPlugType; }
      set {
        if (panelPlugType == value) {
          return;
        }
        if (layoutSlot != null) {
          if (panelPlugType == null) {
            Runtime.Repository.PluginAdded -= Repository_PluginAdded;
          } else {
            Deactivate();
          }
        }
        panelPlugType = value;
        if (layoutSlot != null) {
          if (panelPlugType == null) {
            Runtime.Repository.PluginAdded += Repository_PluginAdded;
          } else {
            Activate();
          }
        }
      }
    }

    private void Repository_PluginAdded(object sender, PluginEventArgs args) {
      PanelPlugType = FindPanelPlugType(
          from extensionType in args.ExtensionTypeInfos
          from plugType in extensionType.PlugTypeInfos
          select plugType
          );
    }

    private void Activate() {
      layoutSlot.Registered += RegisterElement;
      layoutSlot.Deregistered += DeregisterElement;
      if (!layoutSlot.IsOpen) {
        layoutSlot.Open();
        layoutSlot.Register(panelPlugType);
      } else {

        foreach (PlugTypeInfo plugType in layoutSlot.RegisteredPlugTypeInfos) {
          RegisterElement(plugType);
        }
      }
    }

    private void Deactivate() {
      layoutSlot.Unplug();
      foreach (ILayoutPanel layoutPanel in layoutPanels.ToArray()) {
        RemoveLayoutPanel(layoutPanel);
      }
      layoutSlot.Registered -= RegisterElement;
      layoutSlot.Deregistered -= DeregisterElement;
    }

    private PlugTypeInfo FindPanelPlugType(IEnumerable<PlugTypeInfo> plugTypes) {
      //if (layoutSlot == null) {
      //  return null;
      //}
      foreach (PlugTypeInfo plugType in plugTypes.
          Where(plugType => layoutSlot.Name == plugType.Name).
          Where(plugType => plugType.ExtensionTypeInfo.Type.GetInterface(typeof(ILayoutPanel).FullName) != null)) {
        if (!string.IsNullOrEmpty(LayoutPanelTag)) {
          string tag = plugType.GetParamValue("Tag") as string;
          if (string.IsNullOrEmpty(tag) || !LayoutPanelTag.Equals(tag)) {
            continue;
          }
        }
        return plugType;
      }
      return null;
    }

    public ExtensionInfo AddPanel(PanelInfo panelInfo, LayoutInfo layoutInfo) {
      ILayoutPanel panel = (ILayoutPanel) panelPlugType.ExtensionTypeInfo.CreateUniqueExtension().Object;
      panel.PanelInfo = panelInfo;
      panel.LayoutInfo = layoutInfo;
      layoutPanels.Add(panel);
      layoutInfos.Add(panel.ElementPlug, layoutInfo);
      panel.ElementSlot.Plugged += UpdateLayoutInfo;
      panel.ExtensionInfo.Released += RemoveLayoutPanel;
      PlugLayoutPanel(panel);
      return panel.ExtensionInfo;
    }

    private void RemoveLayoutPanel(object sender, ExtensionEventArgs args) {
      RemoveLayoutPanel((ILayoutPanel) args.Extension);
    }

    private void RemoveLayoutPanel(ILayoutPanel layoutPanel) {
      layoutPanel.ElementSlot.Plugged -= UpdateLayoutInfo;
      layoutPanel.ExtensionInfo.Released -= RemoveLayoutPanel;
      foreach (PlugInfo plugInfo in layoutPanel.ElementSlot.PluggedPlugInfos) {
        plugInfo.ExtensionInfo.Released -= RemoveLayoutInfo;
        RemoveLayoutInfo(plugInfo);
      }
      layoutPanels.Remove(layoutPanel);
      layoutInfos.Remove(layoutPanel.ElementPlug);
    }

    private void PlugLayoutPanel(ILayoutPanel panel) {
      ILayoutPanel hostPanel = FindHostPanel(panel);
      SlotInfo panelSlot = panel.ElementSlot;
      SlotInfo hostSlot = hostPanel == null ? layoutSlot : hostPanel.ElementSlot;
      panelSlot.Open();
      panelSlot.Register(panel.ElementPlug.PlugTypeInfo);
      foreach (PlugInfo plug in hostSlot.PluggedPlugInfos) {
        if (IsExtensionNestedPanel(panel, plug.ExtensionInfo)) {
          hostSlot.Unplug(plug);
          panelSlot.Plug(plug);
        }
      }
      hostSlot.Plug(panel.ElementPlug);
    }

    private ILayoutPanel FindHostPanel(ILayoutPanel layoutPanel) {
      ILayoutPanel host = null;
      foreach (ILayoutPanel panel in layoutPanels) {
        if (layoutPanel.LayoutInfo.NestingLevel <= panel.LayoutInfo.NestingLevel) {
          continue;
        }
        if (host == null || host.LayoutInfo.NestingLevel < panel.LayoutInfo.NestingLevel) {
          host = panel;
        }
      }
      return host;
    }

    private static bool IsExtensionNestedPanel(ILayoutPanel layoutPanel, ExtensionInfo extension) {
      return (extension.Object is ILayoutPanel) &&
          ((ILayoutPanel) extension.Object).LayoutInfo.NestingLevel > layoutPanel.LayoutInfo.NestingLevel;
    }

    private void UpdateLayoutInfo(object sender, PlugEventArgs args) {
      //IElement element = (IElement) args.Extension;
      //LayoutPanel hostPanel = (LayoutPanel) args.SlotInfo.ExtensionInfo.Object;
      //hostPanel.AddElement(element.Element, layoutInfos[element.Element]);
    }

    private ILayoutPanel GetPanel(LayoutInfo layoutInfo) {
      foreach (ILayoutPanel panel in layoutPanels) {
        if (panel.PanelInfo.Name != layoutInfo.Panel) {
          continue;
        }
        return panel;
      }
      foreach (ILayoutPanel panel in layoutPanels) {
        if (panel.LayoutInfo.NestingLevel != layoutInfo.NestingLevel) {
          continue;
        }
        return panel;
      }
      ExtensionInfo panelExtension = AddPanel(
          new PanelInfo(layoutInfo.Panel),
          new LayoutInfo(layoutInfo.NestingLevel, 0.5, SizeMode.Percent, 1));
      return (ILayoutPanel) panelExtension.Object;
    }

    private void RegisterElement(object sender, RegisterEventArgs args) {
      RegisterElement(args.PlugTypeInfo);
    }

    private void RegisterElement(PlugTypeInfo plugType) {
      if (panelPlugType == null) {
        return;
      }
      if (!string.IsNullOrEmpty(ElementTag)) {
        string tag = plugType.GetParamValue<string>("Tag");
        if (tag != ElementTag) {
          return;
        }
      }
      LayoutInfo layoutInfo = ParseLayoutInfo(plugType, OnLogEvent);
      ILayoutPanel hostPanel = GetPanel(layoutInfo);
      ExtensionInfo extension = plugType.ExtensionTypeInfo.CreateUniqueExtension();
      extension.Released += RemoveLayoutInfo;
      AddLayoutInfo(extension.PlugInfos[plugType.Name], layoutInfo);
      SlotInfo hostSlot = hostPanel.ElementSlot;
      hostSlot.Register(plugType);
      hostSlot.Plug(extension.PlugInfos[plugType.Name]);
    }

    private void DeregisterElement(object sender, RegisterEventArgs args) {
      DeregisterElement(args.PlugTypeInfo);
    }

    private void DeregisterElement(PlugTypeInfo plugType) {
    }

    public static LayoutInfo ParseLayoutInfo(PlugTypeInfo plugType, LogEventHandler logHandler) {
      string layoutString = plugType.GetParamValue<string>("Layout");
      if (String.IsNullOrEmpty(layoutString)) {
        return null;
      }
      string panelName = string.Empty;
      double nestingLevel = 0;
      double order = 0;
      SizeMode sizeMode = SizeMode.Auto;
      double size = 0;
      string[] layoutParts = layoutString.Split(';');

      foreach (string layoutPart in layoutParts) {
        string[] keyValuePair = layoutPart.Split('=');
        if (keyValuePair.Length != 2) {
          LogParseError(logHandler, plugType, "Invalid layout description: {0}", layoutPart);
          continue;
        }
        string key = keyValuePair[0].Trim();
        string value = keyValuePair[1].Trim();
        switch (key) {
          case "Panel":
            panelName = value;
            break;
          case "NestingLevel":
            if (!double.TryParse(value, NumberStyles.Float, CultureInfo.InvariantCulture, out nestingLevel)) {
              LogParseError(logHandler, plugType, "Invalid nesting level: {0}", value);
            }
            break;
          case "Order":
            if (!double.TryParse(value, NumberStyles.Float, CultureInfo.InvariantCulture, out order)) {
              LogParseError(logHandler, plugType, "Invalid order: {0}", value);
            }
            break;
          case "SizeMode":
            bool invalidSizeMode;
            try {
              sizeMode = (SizeMode) Enum.Parse(typeof(SizeMode), value);
              invalidSizeMode = !Enum.IsDefined(typeof(SizeMode), sizeMode);
            } catch {
              invalidSizeMode = true;
            }
            if (invalidSizeMode) {
              sizeMode = SizeMode.Auto;
              LogParseError(logHandler, plugType, "Invalid size mode: {0}", value);
            }
            break;
          case "Size":
            if (!double.TryParse(value, NumberStyles.Float, CultureInfo.InvariantCulture, out size)) {
              LogParseError(logHandler, plugType, "Invalid size: {0}", value);
            }
            break;
        }
      }
      return new LayoutInfo(panelName, nestingLevel, order, sizeMode, size);
    }

    private static void LogParseError(LogEventHandler logHandler, PlugTypeInfo plugType, string message, params string[] args) {
      if (logHandler == null) {
        return;
      }
      string log = string.Format("Error on parsing layout parameter for plug type {0}:{1}\n{2}",
          plugType.Id, plugType.Name, string.Format(message, args));
      logHandler(null, new LogEventArgs(MessageType.Error, log));
    }

    public static LayoutInfo GetLayoutInfo(PlugInfo plug) {
      return layoutInfos.ContainsKey(plug) ? layoutInfos[plug] : null;
    }

    public static void AddLayoutInfo(PlugInfo plug, LayoutInfo layoutInfo) {
      layoutInfos[plug] = layoutInfo;
    }

    public void RemoveLayoutInfo(object sender, ExtensionEventArgs args) {
      RemoveLayoutInfo(args.ExtensionInfo.PlugInfos[layoutSlot.Name]);
    }

    public static void RemoveLayoutInfo(PlugInfo plug) {
      layoutInfos.Remove(plug);
    }

    private void OnLogEvent(object sender, LogEventArgs args) {
      Extension.LogEvent(args);
    }
  }
}