﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Threading;

namespace Plux.Framework {
  public class ScriptRunner {
    private readonly LogEventHandler log;
    // queue of scripts which will be executed in the script thread
    private Queue<ScriptEntry> scriptQueue = new Queue<ScriptEntry>();
    private System.Threading.Thread scriptThread = null;

    public ScriptRunner(IScriptExecutor scriptExecutor, LogEventHandler logHandler) {
      log = logHandler;
      ScriptExecutor = scriptExecutor;
      scriptThread = new System.Threading.Thread(new ThreadStart(RunScriptsThread));
    }

    public ScriptRunner(LogEventHandler log) {
      this.log = log;
    }

    public IScriptExecutor ScriptExecutor { get; set; }

    public void ExecuteStartupScripts() {
      if (!Runtime.Arguments.ContainsKey("scripts")) {
        LogEvent(MessageType.Error, "Error on executing startup scripts:\n"
            + "  Script argument not found. "
                + "Please use \"/args:scripts=<path1>[,<path2>,...]\" argument for startup");
        return;
      }
      string scriptFileString = Runtime.Arguments["scripts"];
      if (String.IsNullOrEmpty(scriptFileString)) {
        LogEvent(MessageType.Error, "Error on executing startup scripts:\n"
            + "  Script argument must not be emtpy");
        return;
      }
      string[] scriptFiles = scriptFileString.Split(new[] {','}, StringSplitOptions.RemoveEmptyEntries);
      foreach (string scriptFile in scriptFiles) {
        ExecuteScriptFile(scriptFile);
      }
    }

    public void ExecuteScriptFile(string path) {
      if (!File.Exists(path)) {
        LogEvent(
            MessageType.Error, "Error on executing script file: \"{0}\"\n"
                + "  Script file not found", path);
        return;
      }
      string script;
      try {
        script = File.ReadAllText(path);
      } catch (Exception) {
        LogEvent(MessageType.Error,
            "Unknown error on reading script file: \"{0}\"", path);
        return;
      }
      if (script.Length == 0) {
        LogEvent(MessageType.Warning, "Warning on executing script file: {0}\n"
            + "  Script file is empty", path);
        return;
      }
      if (ScriptExecutor == null) {
        LogEvent(MessageType.Error, "Error on executing script file: \"{0}\"\n"
            + "  Script executor is missing", path);
        return;
      }
      InternalExecuteScript(path,script);
    }

    public void ExecuteScriptResource(PluginInfo pluginInfo, string scriptName) {
      using (ResourceStreamGetter rsg =
          new ResourceStreamGetter(pluginInfo, scriptName, false)) {
        using (StreamReader r = new StreamReader(rsg.Stream)) {
          try {
            string script = r.ReadToEnd();
            if (ScriptExecutor == null) {
              LogEvent(MessageType.Error, "Error on executing script resource: \"{0}-{1}\"\n"
                  + "  Script executor is missing", pluginInfo, scriptName);
              return;
            }
            InternalExecuteScript(scriptName,script);
          } catch (Exception e) {
            LogEvent(MessageType.Error,
                "Unknown error on reading script resource: \"{0}-{1}\"\n  {2}",
                pluginInfo.Name, scriptName, e.Message);
          }
        }
      }
    }

    public void ExecuteScript(string script) {
        string identifier = script;
        if (identifier.Length > 40)
            identifier = identifier.Substring(0, 37) + "...";
      InternalExecuteScript(identifier,script);
    }

    private void LogScriptResult(IScriptResult res)
    {
        LogEvent(MessageType.Info, "Script Result: StateMessage: {0}, Errors: {1}.",res.StateMessage,res.Errors.Length);
        if (res.Errors.Length > 0 || res.State != CommandState.Completed)
        {
            // Exception: Input String not in correct format
            //LogEvent(MessageType.Error, res.ToString());
        }
    }

    private void RunScriptsThread()
    {
        bool workOn = true;
        do
        {
            ScriptEntry script;
            lock (scriptQueue)
            {
                script = scriptQueue.Dequeue();
                if (scriptQueue.Count == 0)
                    workOn = false;
            }
            LogEvent(MessageType.Info, "Executing (ScriptThread): " + script.Identifier);
            IScriptResult res = ScriptExecutor.ExecuteSync(script.Script);
            LogScriptResult(res);
        }
        while (workOn);
        LogEvent(MessageType.Info, "Script thread executed all queued scripts.");
    }

    private void InternalExecuteScript(string identifier, string script) {

        if (ScriptExecutor == null)
        {
            LogEvent(MessageType.Error, "Error on executing script: Script executor is missing");
            return;
        }

      if (Runtime.IsRuntimeThread())
      {
          LogEvent(MessageType.Info, "Enqueuing: " + identifier);
          lock (scriptQueue)
          {
              scriptQueue.Enqueue(new ScriptEntry(identifier,script));
              if (!scriptThread.IsAlive)
              {
                  LogEvent(MessageType.Info, "Starting script thread.");
                  scriptThread.Start();
              }
          }
      }
      else
      {
          // execute Immediate
          LogEvent(MessageType.Info, "Executing immediate: " + identifier);
          LogScriptResult(ScriptExecutor.ExecuteSync(script));
      }
    }

    private void LogEvent(MessageType type, string message, params object[] args) {
      log(this, new LogEventArgs(type, VerbosityLevel.Normal, message, args));
    }

    class ScriptEntry
    {
        public string Identifier;
        public string Script;

        public ScriptEntry(string identifier, string script)
        {
            this.Identifier = identifier;
            this.Script = script;
        }
    }
  }
}